from typing import Any

def is_html_input(dictionary: Any): ...
def parse_html_list(dictionary: Any, prefix: str = ..., default: Any | None = ...): ...
def parse_html_dict(dictionary: Any, prefix: str = ...): ...
